@@ -3,25 +3,44 @@ Item Loader
 
 See documentation in docs/topics/loaders.rst
 """
-from collections import defaultdict
+
+from __future__ import annotations
+
 from contextlib import suppress
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    MutableMapping,
+    Optional,
+    Pattern,
+    Union,
+)
 
 from itemadapter import ItemAdapter
+from parsel import Selector
 from parsel.utils import extract_regex, flatten
 
 from itemloaders.common import wrap_loader_context
 from itemloaders.processors import Identity
 from itemloaders.utils import arg_to_iter
 
+if TYPE_CHECKING:
+    # typing.Self requires Python 3.11
+    from typing_extensions import Self
 
-def unbound_method(method):
+
+def unbound_method(method: Callable[..., Any]) -> Callable[..., Any]:
     """
     Allow to use single-argument functions as input or output processors
     (no need to define an unused first 'self' argument)
     """
     with suppress(AttributeError):
-        if '.' not in method.__qualname__:
-            return method.__func__
+        if "." not in method.__qualname__:
+            return method.__func__  # type: ignore[attr-defined, no-any-return]
     return method
 
 
@@ -37,13 +56,13 @@ class ItemLoader:
 
     :param item: The item instance to populate using subsequent calls to
         :meth:`~ItemLoader.add_xpath`, :meth:`~ItemLoader.add_css`,
-        or :meth:`~ItemLoader.add_value`.
+        :meth:`~ItemLoader.add_jmes` or :meth:`~ItemLoader.add_value`.
     :type item: :class:`dict` object
 
     :param selector: The selector to extract data from, when using the
-        :meth:`add_xpath` (resp. :meth:`add_css`) or :meth:`replace_xpath`
-        (resp. :meth:`replace_css`) method.
-    :type selector: :class:`~parsel.Selector` object
+        :meth:`add_xpath` (resp. :meth:`add_css`, :meth:`add_jmes`) or :meth:`replace_xpath`
+        (resp. :meth:`replace_css`, :meth:`replace_jmes`) method.
+    :type selector: :class:`~parsel.selector.Selector` object
 
     The item, selector and the remaining keyword arguments are
     assigned to the Loader context (accessible through the :attr:`context` attribute).
@@ -89,46 +108,53 @@ class ItemLoader:
 
     .. attribute:: selector
 
-        The :class:`~parsel.Selector` object to extract data from.
+        The :class:`~parsel.selector.Selector` object to extract data from.
         It's the selector given in the ``__init__`` method.
         This attribute is meant to be read-only.
 
     .. _parsel: https://parsel.readthedocs.io/en/latest/
     """
 
-    default_item_class = dict
-    default_input_processor = Identity()
-    default_output_processor = Identity()
+    default_item_class: type = dict
+    default_input_processor: Callable[..., Any] = Identity()
+    default_output_processor: Callable[..., Any] = Identity()
 
-    def __init__(self, item=None, selector=None, parent=None, **context):
-        self.selector = selector
+    def __init__(
+        self,
+        item: Any = None,
+        selector: Optional[Selector] = None,
+        parent: Optional[ItemLoader] = None,
+        **context: Any,
+    ):
+        self.selector: Optional[Selector] = selector
         context.update(selector=selector)
         if item is None:
             item = self.default_item_class()
         self._local_item = item
-        context['item'] = item
-        self.context = context
-        self.parent = parent
-        self._local_values = defaultdict(list)
+        context["item"] = item
+        self.context: MutableMapping[str, Any] = context
+        self.parent: Optional[ItemLoader] = parent
+        self._local_values: Dict[str, List[Any]] = {}
         # values from initial item
         for field_name, value in ItemAdapter(item).items():
+            self._values.setdefault(field_name, [])
             self._values[field_name] += arg_to_iter(value)
 
     @property
-    def _values(self):
+    def _values(self) -> Dict[str, List[Any]]:
         if self.parent is not None:
             return self.parent._values
         else:
             return self._local_values
 
     @property
-    def item(self):
+    def item(self) -> Any:
         if self.parent is not None:
             return self.parent.item
         else:
             return self._local_item
 
-    def nested_xpath(self, xpath, **context):
+    def nested_xpath(self, xpath: str, **context: Any) -> Self:
         """
         Create a nested loader with an xpath selector.
         The supplied selector is applied relative to selector associated
@@ -136,14 +162,14 @@ class ItemLoader:
         with the parent :class:`ItemLoader` so calls to :meth:`add_xpath`,
         :meth:`add_value`, :meth:`replace_value`, etc. will behave as expected.
         """
+        self._check_selector_method()
+        assert self.selector is not None
         selector = self.selector.xpath(xpath)
         context.update(selector=selector)
-        subloader = self.__class__(
-            item=self.item, parent=self, **context
-        )
+        subloader = self.__class__(item=self.item, parent=self, **context)
         return subloader
 
-    def nested_css(self, css, **context):
+    def nested_css(self, css: str, **context: Any) -> Self:
         """
         Create a nested loader with a css selector.
         The supplied selector is applied relative to selector associated
@@ -151,14 +177,21 @@ class ItemLoader:
         with the parent :class:`ItemLoader` so calls to :meth:`add_xpath`,
         :meth:`add_value`, :meth:`replace_value`, etc. will behave as expected.
         """
+        self._check_selector_method()
+        assert self.selector is not None
         selector = self.selector.css(css)
         context.update(selector=selector)
-        subloader = self.__class__(
-            item=self.item, parent=self, **context
-        )
+        subloader = self.__class__(item=self.item, parent=self, **context)
         return subloader
 
-    def add_value(self, field_name, value, *processors, **kw):
+    def add_value(
+        self,
+        field_name: Optional[str],
+        value: Any,
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Self:
         """
         Process and then add the given ``value`` for the given field.
 
@@ -172,6 +205,9 @@ class ItemLoader:
         multiple fields may be added. And the processed value should be a dict
         with field_name mapped to values.
 
+        :returns: The current ItemLoader instance for method chaining.
+        :rtype: ItemLoader
+
         Examples::
 
             loader.add_value('name', 'Color TV')
@@ -179,41 +215,61 @@ class ItemLoader:
             loader.add_value('length', '100')
             loader.add_value('name', 'name: foo', TakeFirst(), re='name: (.+)')
             loader.add_value(None, {'name': 'foo', 'sex': 'male'})
+
         """
-        value = self.get_value(value, *processors, **kw)
+        value = self.get_value(value, *processors, re=re, **kw)
         if value is None:
-            return
+            return self
         if not field_name:
             for k, v in value.items():
                 self._add_value(k, v)
         else:
             self._add_value(field_name, value)
+        return self
 
-    def replace_value(self, field_name, value, *processors, **kw):
+    def replace_value(
+        self,
+        field_name: Optional[str],
+        value: Any,
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Self:
         """
         Similar to :meth:`add_value` but replaces the collected data with the
         new value instead of adding it.
+
+        :returns: The current ItemLoader instance for method chaining.
+        :rtype: ItemLoader
         """
-        value = self.get_value(value, *processors, **kw)
+        value = self.get_value(value, *processors, re=re, **kw)
         if value is None:
-            return
+            return self
         if not field_name:
             for k, v in value.items():
                 self._replace_value(k, v)
         else:
             self._replace_value(field_name, value)
+        return self
 
-    def _add_value(self, field_name, value):
+    def _add_value(self, field_name: str, value: Any) -> None:
         value = arg_to_iter(value)
         processed_value = self._process_input_value(field_name, value)
         if processed_value:
+            self._values.setdefault(field_name, [])
             self._values[field_name] += arg_to_iter(processed_value)
 
-    def _replace_value(self, field_name, value):
+    def _replace_value(self, field_name: str, value: Any) -> None:
         self._values.pop(field_name, None)
         self._add_value(field_name, value)
 
-    def get_value(self, value, *processors, **kw):
+    def get_value(
+        self,
+        value: Any,
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Any:
         """
         Process the given ``value`` by the given ``processors`` and keyword
         arguments.
@@ -221,9 +277,9 @@ class ItemLoader:
         Available keyword arguments:
 
         :param re: a regular expression to use for extracting data from the
-            given value using :meth:`~parsel.utils.extract_regex` method,
+            given value using :func:`~parsel.utils.extract_regex` method,
             applied before processors
-        :type re: str or compiled regex
+        :type re: str or typing.Pattern[str]
 
         Examples:
 
@@ -233,10 +289,9 @@ class ItemLoader:
         >>> loader.get_value('name: foo', TakeFirst(), str.upper, re='name: (.+)')
         'FOO'
         """
-        regex = kw.get('re', None)
-        if regex:
+        if re:
             value = arg_to_iter(value)
-            value = flatten(extract_regex(regex, x) for x in value)
+            value = flatten(extract_regex(re, x) for x in value)
 
         for proc in processors:
             if value is None:
@@ -246,12 +301,13 @@ class ItemLoader:
             try:
                 value = proc(value)
             except Exception as e:
-                raise ValueError("Error with processor %s value=%r error='%s: %s'" %
-                                 (_proc.__class__.__name__, value,
-                                  type(e).__name__, str(e)))
+                raise ValueError(
+                    "Error with processor %s value=%r error='%s: %s'"
+                    % (_proc.__class__.__name__, value, type(e).__name__, str(e))
+                ) from e
         return value
 
-    def load_item(self):
+    def load_item(self) -> Any:
         """
         Populate the item with the data collected so far, and return it. The
         data collected is first passed through the :ref:`output processors
@@ -265,48 +321,49 @@ class ItemLoader:
 
         return adapter.item
 
-    def get_output_value(self, field_name):
+    def get_output_value(self, field_name: str) -> Any:
         """
         Return the collected values parsed using the output processor, for the
         given field. This method doesn't populate or modify the item at all.
         """
         proc = self.get_output_processor(field_name)
         proc = wrap_loader_context(proc, self.context)
+        value = self._values.get(field_name, [])
         try:
-            return proc(self._values[field_name])
+            return proc(value)
         except Exception as e:
-            raise ValueError("Error with output processor: field=%r value=%r error='%s: %s'" %
-                             (field_name, self._values[field_name], type(e).__name__, str(e)))
+            raise ValueError(
+                "Error with output processor: field=%r value=%r error='%s: %s'"
+                % (field_name, value, type(e).__name__, str(e))
+            ) from e
 
-    def get_collected_values(self, field_name):
+    def get_collected_values(self, field_name: str) -> List[Any]:
         """Return the collected values for the given field."""
-        return self._values[field_name]
+        return self._values.get(field_name, [])
 
-    def get_input_processor(self, field_name):
-        proc = getattr(self, '%s_in' % field_name, None)
+    def get_input_processor(self, field_name: str) -> Callable[..., Any]:
+        proc = getattr(self, "%s_in" % field_name, None)
         if not proc:
             proc = self._get_item_field_attr(
-                field_name,
-                'input_processor',
-                self.default_input_processor
+                field_name, "input_processor", self.default_input_processor
             )
         return unbound_method(proc)
 
-    def get_output_processor(self, field_name):
-        proc = getattr(self, '%s_out' % field_name, None)
+    def get_output_processor(self, field_name: str) -> Callable[..., Any]:
+        proc = getattr(self, "%s_out" % field_name, None)
         if not proc:
             proc = self._get_item_field_attr(
-                field_name,
-                'output_processor',
-                self.default_output_processor
+                field_name, "output_processor", self.default_output_processor
             )
         return unbound_method(proc)
 
-    def _get_item_field_attr(self, field_name, key, default=None):
+    def _get_item_field_attr(
+        self, field_name: str, key: Any, default: Any = None
+    ) -> Any:
         field_meta = ItemAdapter(self.item).get_field_meta(field_name)
         return field_meta.get(key, default)
 
-    def _process_input_value(self, field_name, value):
+    def _process_input_value(self, field_name: str, value: Any) -> Any:
         proc = self.get_input_processor(field_name)
         _proc = proc
         proc = wrap_loader_context(proc, self.context)
@@ -315,17 +372,31 @@ class ItemLoader:
         except Exception as e:
             raise ValueError(
                 "Error with input processor %s: field=%r value=%r "
-                "error='%s: %s'" % (_proc.__class__.__name__, field_name,
-                                    value, type(e).__name__, str(e)))
+                "error='%s: %s'"
+                % (
+                    _proc.__class__.__name__,
+                    field_name,
+                    value,
+                    type(e).__name__,
+                    str(e),
+                )
+            ) from e
 
-    def _check_selector_method(self):
+    def _check_selector_method(self) -> None:
         if self.selector is None:
             raise RuntimeError(
-                "To use XPath or CSS selectors, %s"
+                "To use XPath or CSS selectors, %s "
                 "must be instantiated with a selector" % self.__class__.__name__
             )
 
-    def add_xpath(self, field_name, xpath, *processors, **kw):
+    def add_xpath(
+        self,
+        field_name: Optional[str],
+        xpath: Union[str, Iterable[str]],
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Self:
         """
         Similar to :meth:`ItemLoader.add_value` but receives an XPath instead of a
         value, which is used to extract a list of strings from the
@@ -336,6 +407,9 @@ class ItemLoader:
         :param xpath: the XPath to extract data from
         :type xpath: str
 
+        :returns: The current ItemLoader instance for method chaining.
+        :rtype: ItemLoader
+
         Examples::
 
             # HTML snippet: <p class="product-name">Color TV</p>
@@ -345,16 +419,33 @@ class ItemLoader:
 
         """
         values = self._get_xpathvalues(xpath, **kw)
-        self.add_value(field_name, values, *processors, **kw)
+        return self.add_value(field_name, values, *processors, re=re, **kw)
 
-    def replace_xpath(self, field_name, xpath, *processors, **kw):
+    def replace_xpath(
+        self,
+        field_name: Optional[str],
+        xpath: Union[str, Iterable[str]],
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Self:
         """
         Similar to :meth:`add_xpath` but replaces collected data instead of adding it.
+
+        :returns: The current ItemLoader instance for method chaining.
+        :rtype: ItemLoader
+
         """
         values = self._get_xpathvalues(xpath, **kw)
-        self.replace_value(field_name, values, *processors, **kw)
+        return self.replace_value(field_name, values, *processors, re=re, **kw)
 
-    def get_xpath(self, xpath, *processors, **kw):
+    def get_xpath(
+        self,
+        xpath: Union[str, Iterable[str]],
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Any:
         """
         Similar to :meth:`ItemLoader.get_value` but receives an XPath instead of a
         value, which is used to extract a list of unicode strings from the
@@ -365,7 +456,7 @@ class ItemLoader:
 
         :param re: a regular expression to use for extracting data from the
             selected XPath region
-        :type re: str or compiled regex
+        :type re: str or typing.Pattern[str]
 
         Examples::
 
@@ -376,14 +467,24 @@ class ItemLoader:
 
         """
         values = self._get_xpathvalues(xpath, **kw)
-        return self.get_value(values, *processors, **kw)
+        return self.get_value(values, *processors, re=re, **kw)
 
-    def _get_xpathvalues(self, xpaths, **kw):
+    def _get_xpathvalues(
+        self, xpaths: Union[str, Iterable[str]], **kw: Any
+    ) -> List[Any]:
         self._check_selector_method()
+        assert self.selector is not None
         xpaths = arg_to_iter(xpaths)
-        return flatten(self.selector.xpath(xpath).getall() for xpath in xpaths)
+        return flatten(self.selector.xpath(xpath, **kw).getall() for xpath in xpaths)
 
-    def add_css(self, field_name, css, *processors, **kw):
+    def add_css(
+        self,
+        field_name: Optional[str],
+        css: Union[str, Iterable[str]],
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Self:
         """
         Similar to :meth:`ItemLoader.add_value` but receives a CSS selector
         instead of a value, which is used to extract a list of unicode strings
@@ -394,24 +495,45 @@ class ItemLoader:
         :param css: the CSS selector to extract data from
         :type css: str
 
+        :returns: The current ItemLoader instance for method chaining.
+        :rtype: ItemLoader
+
         Examples::
 
             # HTML snippet: <p class="product-name">Color TV</p>
             loader.add_css('name', 'p.product-name')
             # HTML snippet: <p id="price">the price is $1200</p>
             loader.add_css('price', 'p#price', re='the price is (.*)')
-        """
-        values = self._get_cssvalues(css, **kw)
-        self.add_value(field_name, values, *processors, **kw)
 
-    def replace_css(self, field_name, css, *processors, **kw):
+        """
+        values = self._get_cssvalues(css)
+        return self.add_value(field_name, values, *processors, re=re, **kw)
+
+    def replace_css(
+        self,
+        field_name: Optional[str],
+        css: Union[str, Iterable[str]],
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Self:
         """
         Similar to :meth:`add_css` but replaces collected data instead of adding it.
-        """
-        values = self._get_cssvalues(css, **kw)
-        self.replace_value(field_name, values, *processors, **kw)
 
-    def get_css(self, css, *processors, **kw):
+        :returns: The current ItemLoader instance for method chaining.
+        :rtype: ItemLoader
+
+        """
+        values = self._get_cssvalues(css)
+        return self.replace_value(field_name, values, *processors, re=re, **kw)
+
+    def get_css(
+        self,
+        css: Union[str, Iterable[str]],
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Any:
         """
         Similar to :meth:`ItemLoader.get_value` but receives a CSS selector
         instead of a value, which is used to extract a list of unicode strings
@@ -422,7 +544,7 @@ class ItemLoader:
 
         :param re: a regular expression to use for extracting data from the
             selected CSS region
-        :type re: str or compiled regex
+        :type re: str or typing.Pattern[str]
 
         Examples::
 
@@ -431,10 +553,98 @@ class ItemLoader:
             # HTML snippet: <p id="price">the price is $1200</p>
             loader.get_css('p#price', TakeFirst(), re='the price is (.*)')
         """
-        values = self._get_cssvalues(css, **kw)
-        return self.get_value(values, *processors, **kw)
+        values = self._get_cssvalues(css)
+        return self.get_value(values, *processors, re=re, **kw)
 
-    def _get_cssvalues(self, csss, **kw):
+    def _get_cssvalues(self, csss: Union[str, Iterable[str]]) -> List[Any]:
         self._check_selector_method()
+        assert self.selector is not None
         csss = arg_to_iter(csss)
         return flatten(self.selector.css(css).getall() for css in csss)
+
+    def add_jmes(
+        self,
+        field_name: Optional[str],
+        jmes: str,
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Self:
+        """
+        Similar to :meth:`ItemLoader.add_value` but receives a JMESPath selector
+        instead of a value, which is used to extract a list of unicode strings
+        from the selector associated with this :class:`ItemLoader`.
+
+        See :meth:`get_jmes` for ``kwargs``.
+
+        :param jmes: the JMESPath selector to extract data from
+        :type jmes: str
+
+        :returns: The current ItemLoader instance for method chaining.
+        :rtype: ItemLoader
+
+        Examples::
+
+            # HTML snippet: {"name": "Color TV"}
+            loader.add_jmes('name')
+            # HTML snippet: {"price": the price is $1200"}
+            loader.add_jmes('price', TakeFirst(), re='the price is (.*)')
+        """
+        values = self._get_jmesvalues(jmes)
+        return self.add_value(field_name, values, *processors, re=re, **kw)
+
+    def replace_jmes(
+        self,
+        field_name: Optional[str],
+        jmes: Union[str, Iterable[str]],
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Self:
+        """
+        Similar to :meth:`add_jmes` but replaces collected data instead of adding it.
+
+        :returns: The current ItemLoader instance for method chaining.
+        :rtype: ItemLoader
+        """
+        values = self._get_jmesvalues(jmes)
+        return self.replace_value(field_name, values, *processors, re=re, **kw)
+
+    def get_jmes(
+        self,
+        jmes: Union[str, Iterable[str]],
+        *processors: Callable[..., Any],
+        re: Union[str, Pattern[str], None] = None,
+        **kw: Any,
+    ) -> Any:
+        """
+        Similar to :meth:`ItemLoader.get_value` but receives a JMESPath selector
+        instead of a value, which is used to extract a list of unicode strings
+        from the selector associated with this :class:`ItemLoader`.
+
+        :param jmes: the JMESPath selector to extract data from
+        :type jmes: str
+
+        :param re: a regular expression to use for extracting data from the
+            selected JMESPath
+        :type re: str or typing.Pattern
+
+        Examples::
+
+            # HTML snippet: {"name": "Color TV"}
+            loader.get_jmes('name')
+            # HTML snippet: {"price": the price is $1200"}
+            loader.get_jmes('price', TakeFirst(), re='the price is (.*)')
+        """
+        values = self._get_jmesvalues(jmes)
+        return self.get_value(values, *processors, re=re, **kw)
+
+    def _get_jmesvalues(self, jmess: Union[str, Iterable[str]]) -> List[Any]:
+        self._check_selector_method()
+        assert self.selector is not None
+        jmess = arg_to_iter(jmess)
+        if not hasattr(self.selector, "jmespath"):
+            raise AttributeError(
+                "Please install parsel >= 1.8.1 to get jmespath support"
+            )
+        return flatten(self.selector.jmespath(jmes).getall() for jmes in jmess)
